/* This is a small disassembler written by Bill Rogers. It was later robbed
 * from AmiNet, pillaged and somewhat burnt by Per Haral Myrvang.
 * Note: This code wouldn't be re-entrant in a million years!
 * Note 2: This code is occasionally cursed!!
 */


#include <Pilot.h>
#include <Hardware/Hardware.h>

#include "disasm.h"

/* Distance from beginning of line to operand output */
#define OPSKIP 64
#define OPBUFSIZE 140
#define DISASS_MAXYLINE 12
#define MEMNONIC_XSTART 60

struct instruction {
    unsigned opcode:  4;
    unsigned destreg: 3;
    unsigned extend:  3;
    unsigned srcmode: 3;
    unsigned srcreg:  3;
};

struct shiftbits {
    unsigned opcode:  4;
    unsigned count:   3;
    unsigned direct:  1;
    unsigned optype:  2;
    unsigned opmode:  1;
    unsigned shift:   2;
    unsigned dest:    3;
};

struct testbits {
    unsigned opcode:  4;
    unsigned ccode:   4;
    unsigned offset:  8;
};

static union {
    struct instruction iview;
    struct shiftbits s;
    struct testbits tview;
    short sview;
} cheat;

typedef struct {
    unsigned char	*address;
    unsigned long	index_into;	/* i.e. relative to address */
    long		c;
    long		codeaddress;	/* keep track of where we are */
    long		codesize;	/* size of decoded operand */
    char		*opp;		/* memnonic */
    char		*valp;		/* hex value */
    char		*opline;	/* buffer */
    int			line;
    int			unreadable;
    struct {
	unsigned do_hexopcodes:	1;	/* Hex AND memnonic disassembly */
	unsigned display_state:	2;	/* disassembly or hexdump */
	unsigned branch_type:	1;	/* relative or absolute address */
	unsigned syscallname:	1;	/* show by name or number */
    } state;
} disass_t;


/* Globals */

static disass_t *fin = NULL;


/*protos*/

static void opchar(register char);
static void operand(long, long);


/* Originally, this was wired so that it would bomb if not called once
 * initially, to check for EOF. Removed for my sanity's sake, if nothing else.
 * In fact, it has been completly rewritten. It now tests to see if the address
 * we try to read from is readable. It also keeps track of which bytes are
 * unreadable, and only throws an error if we're in disassembly mode. This
 * works since fin->unreadable are cleared for every line (i.e. 8 bytes).
 * --Perm */
static long grabchar(void)
{
    unsigned char c;
    if (HwrMemReadable((void *)fin->address+fin->index_into)) {
	c = fin->address[fin->index_into++];
	fin->c = (long)c;
	fin->unreadable <<= 1;
	return fin->c;
    } else {
	fin->index_into++;
	fin->unreadable = (fin->unreadable << 1 ) | 1;
	if (DISP_DISASM == fin->state.display_state)
	    ErrThrow(1);
    }
    return -1;
}

/* Attempt to read 8 bits from address <from>. Returns the byte read, or 0x0100
 * on failure. */
unsigned int
Read_Byte(void *from)
{
    if (HwrMemReadable(from))
	return (unsigned int)(*((unsigned char *)from) & 0x00ff);
    else
	return 0x0100;
}

static long grabword(void)
{
    return (grabchar() << 8 ) | grabchar();
}


#ifdef UNDEF
static long grabthree(void)
{
    return (grabchar() << 16 ) | grabword();
}

static long grablong(void)
{
    return (grabword() << 16 ) | grabword();
}
#endif


static long nextcodeword(void)
{
    fin->codeaddress += 2;
    fin->codesize += 2;
    return grabword();
}


static void startline(void)
{
    fin->valp = fin->opline;
    fin->opp = fin->opline + OPSKIP;
    MemSet(fin->opline, OPBUFSIZE, 0);
}


static void display_hex_line(void)
{
    int i, l, x, s_w;
    char *s, *c, *m;
    for (i = 0; i < OPBUFSIZE; i++)	/* Hack to ensure we print everything. Magic. Don't ask. Go away. */
	if ('\0' == fin->opline[i])
	    fin->opline[i] = ' ';
    fin->opline[OPBUFSIZE] = '\0';
    l = (10 * fin->line++) + LISTING_START_Y + 1;
    c = s = fin->opline;
    x = 0;
    s_w = 15;
    m = fin->address;
    while (*s) {
	while(*c && (' ' != *c))	/* Find space or eol */
	    c++;
	i = c - s;			/* Length of string */
	WinDrawChars(s, i, x, l);	/* Write string */
	x += s_w;
	s_w = 22;
	while (*c && (' ' == *c))	/* Skip any extranous spaces */
	    c++;
	s = c;
    }
    x+= 4;
    WinDrawChars(m, 8, x, l);
}


static void displayline(void)
{
    int i, l, x, w, s_w;
    char *s, *c, *m;

    fin->opline[OPBUFSIZE] = '\0';	/* Just to be sure. */
    l = (10 * fin->line++) + LISTING_START_Y + 1;	/* calc y-coord */
    m = fin->opline + OPSKIP;
    c = s = fin->opline;
    x = 0;
    s_w = 5;
drawtext:
    while (*s) {
	while(*c && (' ' != *c))	/* Find space or eol */
	    c++;
	i = c - s;			/* Length of string */
	w = FntCharsWidth(s, i);	/* Get width of string in pixels */
	if (w + x > 159) {		/* See if it's room */
	    Int nw, nl;
	    Boolean b;
	    nw = 159 - x;		/* Write as much as we can. */
	    nl = i;
	    FntCharsInWidth(s, &nw, &nl, &b);
	    WinDrawChars(s, nl, x, l);
	    WinDrawChars("\x85", 1, 155, l);	/* Append an ellipsis. */
	    s = m;			/* Side-step the goto-hack */
	    break;			/* We're done. */
	}
	WinDrawChars(s, i, x, l);	/* Write string */
	x += w + s_w;			/* A space here is first 5 pixels, */
	s_w = 3;			/* then 3. */
	if (MEMNONIC_XSTART <= x)	/* Attempt to tabularize spaces, */
	    x += (x % s_w);		/* modulo space_width (s_w). */
	while (*c && (' ' == *c))	/* Skip any extranous spaces */
	    c++;
	s = c;
    }
    if (s < m) {			/* If we've only done up to valp */
	c = s = m;			/* do the mnemonic also. */
	x = fin->state.do_hexopcodes ? MEMNONIC_XSTART : MEMNONIC_XSTART - 28;
	goto drawtext;
    }
}


#if 0
static void padd(void)
{
    while (fin->opp < (fin->opline + 22))
	opchar(' ');
}
#else
#define padd() opchar(' ');
#endif


static void opchar(register char c)
{
    *fin->opp++ = c;
}


static void opstring(register char *s)
{
    while (*s != '\0')
	*fin->opp++ = *s++;
}


static void opcode(register char *s)
{
    long n = 5;
    while (*s != '\0' && n-- > 0)
	*fin->opp++ = *s++;
}


static void opint(register long v)
{
    long t;
    if (v <= 0)
	*fin->opp++='0';
    else {
	if ((t = v/10) > 0)
	    opint(t);
	*fin->opp++ = v % 10 + '0';
    };
}


static void ophex(long v, int n)
{
    char c;
    if (n < 0)
	n = -n;
    else 
	opchar('$');
    while (n-- > 0)
	*fin->opp++ = (c = (v >> (n << 2)) & 0xF) <= 9 ? c + '0' : c - 10 + 'A';
    *fin->opp = '\0';
}


static void lefthex(long v, int b, int n)
{
    char c;
    while (b-- > 0)
	*fin->valp++ = ' ';
    while (n-- > 0)
	*fin->valp++ = (c =(v >> (n << 2)) & 0xF) <= 9 ? c + '0' : c - 10 + 'A';
    *fin->valp = '\0';
}


static long mynext(long b)
{
    long v;

    v = nextcodeword();
    if (fin->state.do_hexopcodes)
	lefthex(v, b, 4);
    return v;
}


static void invalid(void)
{
    opstring("dc.w ");
    ophex(cheat.sview, 4);
}


static long srckind;	/* i.e. ".[bwl]" */


static void sourcekind(long k)
{
    opchar('.');
    switch (k) {
	case 0:
	case 4:
	    opchar('b');
	    srckind=0;
	    break;
	case 1:
	case 5:
	case 3:
	    opchar('w');
	    srckind=1;
	    break;
	case 2:
	case 6:
	case 7:
	    opchar('l');
	    srckind=2;
	    break;
    }
}


static void shiftinstruction(void)
{
    long t;
    static char *mnemonics[8] = { "asr",  "asl",  "lsr", "lsl",
	"roxr", "roxl", "ror", "rol"
    };
    if (cheat.s.optype == 3) {
	opstring(mnemonics[(cheat.s.count << 1) + cheat.s.direct]);
	padd();
	operand(cheat.iview.srcmode, cheat.iview.srcreg);
    }
    else {
	opstring(mnemonics[(cheat.s.shift << 1) + cheat.s.direct]);
	sourcekind(cheat.s.optype);
	padd();
	if (cheat.s.opmode == 0) {
	    opchar('#');
	    opint( (t = cheat.s.count) == 0 ? 8 : t );
	}
	else
	    operand(0, cheat.s.count);
	opchar(',');
	operand(0, cheat.s.dest);
    }
}


static void operand(long mode, long reg)
{
    struct indexword {
	unsigned dora :1;
	unsigned ireg :3;
	unsigned size :1;
	unsigned junk :3;
	unsigned disp :8;
    };
    union {
	short sh;
	struct indexword ind;
    } mycheat;
    switch (mode) {
	case 0:
	    opchar('d');
	    opint(reg);
	    break;
	case 1:
	    opchar('a');
	    opint(reg);
	    break;
	case 2:
	    opstring("(a");
	    opint(reg);
	    opchar(')');
	    break;
	case 3:
	    opstring("(a");
	    opint(reg);
	    opstring(")+");
	    break;
	case 4:
	    opstring("-(a");
	    opint(reg);
	    opchar(')');
	    break;
	case 5:
	    ophex(mynext(1), 4);
	    opstring("(a");
	    opint(reg);
	    opchar(')');
	    break;
	case 6:
	    mycheat.sh = mynext(1);
	    ophex(mycheat.ind.disp, 2);
	    opstring("(a");
	    opint(reg);
	    opchar(',');
	    opchar(mycheat.ind.dora == 0 ? 'd' : 'a');
	    opint(mycheat.ind.ireg);
	    opchar('.');
	    opchar(mycheat.ind.size == 0 ? 'w' : 'l');
	    opchar(')');
	    break;
	case 7:
	    switch (reg) {
		case 0:
		    ophex(mynext(1), 4);
		    break;
		case 1:
		    ophex(mynext(1), 4);
		    ophex(mynext(0), -4);
		    break;
		case 2:				/* Also a branch? */
		    if (BRANCH_RELATIVE == fin->state.branch_type) {
			/* relative branch address */
			opchar('+');
			ophex(mynext(1), 4);
		    } else { /* absolute branch address */
			/* minus 2 because mynext() adds 2. */
			ophex(mynext(1) + fin->codeaddress - 2, 8);
		    }
		    break;
		case 3:
		    mycheat.sh = mynext(1);
		    opchar('+');
		    ophex(mycheat.ind.disp, 2);
		    opchar('(');
		    opchar(mycheat.ind.dora == 0 ? 'd' : 'a');
		    opint(mycheat.ind.ireg);
		    opchar('.');
		    opchar(mycheat.ind.size == 0 ? 'w' : 'l');
		    opchar(')');
		    break;
		case 4:
		    opchar('#');
		    switch (srckind) {
			case 0:
			    ophex(mynext(1), 2);
			    break;
			case 1:
			    ophex(mynext(1), 4);
			    break;
			case 2:
			    ophex(mynext(1), 4);
			    ophex(mynext(0), -4); 
			    break;


		    };
		    break;
	    }
    }
}


static void conditioncode(long cc)
{
    char *codes = "rasrhilscccsneeqvcvsplmigeltgtle";
    if (cc > 1 || cheat.iview.opcode == 6) {
	opchar(codes[cc << 1]);
	opchar(codes[(cc << 1) + 1]);
    }
    else
	if (cc == 0)
	    opchar('t');
	else
	    opchar('f');
}


static void handlezero(void)
{
    long ext =  cheat.iview.extend;
    long dreg = cheat.iview.destreg;
    long smod = cheat.iview.srcmode;
    long sreg = cheat.iview.srcreg;
    static char *opzero[] = {"or", "and", "sub", "add", "", "eor", "cmp", ""};
    static char *opone[] = {"btst", "bchg", "bclr", "bset"};

    if (ext <= 2 && dreg != 4) {
	opstring(opzero[dreg]);
	sourcekind(ext);
	padd();
	operand(7,4);
	opchar(',');
	if (smod == 7 && sreg == 4)
	    if (srckind == 0)
		opstring("ccr");
	    else
		opstring("sr");
	else
	    operand(smod,sreg);
    }
    else if (ext <= 3 && dreg == 4) {
	opstring(opone[ext]);
	padd();
	srckind = 0;
	operand(7, 4);
	opchar(',');
	operand(smod,sreg);
    }
    else if (ext >= 4) {
	if (smod != 1) {
	    opstring(opone[ext - 4]);
	    padd();
	    operand(0, dreg);
	    opchar(',');
	    operand(smod, sreg);
	}
	else {
	    opstring("movep");
	    sourcekind(1 + (ext & 1));
	    padd();
	    if (ext <= 5) {
		operand(5, sreg);
		opchar(',');
		operand(0, dreg);
	    }
	    else {
		operand(0, dreg);
		opchar(',');
		operand(5, sreg);
	    }
	}
    }
    else
	invalid();
}


static void breakfurther(char *base)
{
    long ext =  cheat.iview.extend;
    long dreg = cheat.iview.destreg;
    long smod = cheat.iview.srcmode;
    long sreg = cheat.iview.srcreg;
    long comm,t;

    if (((comm = *(base + ext)) & 0xF0 ) == 0x40 )
	comm = *(base + (8 + (t = comm & 0xF) + (t << 2) + (smod == 0 ? 1 : (smod == 1 ? 2 : 0))));
    if (comm==0)
	invalid();
    else {
	opcode(base + (8 + (t = comm & 7) + (t << 2)));
	if (comm & 8)
	    sourcekind(ext);
	if ((t = (comm >> 4) & 0xF) >= 13)	/* branch */
	    conditioncode(cheat.tview.ccode);
	switch (t) {
	    case 0:
		invalid();
		break;
	    case 1:
		padd();
		operand(smod, sreg);
		opchar(',');
		operand(0, dreg);
		break;
	    case 2:
		padd();
		operand(smod, sreg);
		opchar(',');
		operand(1, dreg);
		break;
	    case 3:
		padd();
		operand(0, dreg);
		opchar(',');
		operand(smod,sreg);
		break;
	    case 4:
		opstring(" Internal check fail");
		break;
	    case 5:
		padd();
		operand(3, sreg);
		opchar(',');
		operand(3, dreg);
		break;
	    case 6:
		padd();
		operand(4, sreg);
		opchar(',');
		operand(4, dreg);
		break;
	    case 7:
		opstring(" #");
		opint(dreg);
		opchar(',');
		operand(smod, sreg);
		break;
	    case 8:
		opstring(" #");
		ophex(cheat.tview.offset, 2);
		opchar(',');
		operand(0, dreg);
		break;
	    case 9:
		padd();
		operand(1, dreg);
		opchar(',');
		operand(smod, sreg);
		break;
	    case 13:
		padd();
		operand(smod, sreg);
		break;
	    case 14:
		padd();
		operand(0, sreg);
		opchar(',');
		operand(7, 2);
		break;
	    case 15:		/* branch instruction */
		padd();
		if (cheat.tview.offset == 0)
		    operand(7, 2);
		else {
		    if (BRANCH_RELATIVE == fin->state.branch_type) {
			/* relative branch address */
			opchar('+');
			ophex(cheat.tview.offset, 2);
		    } else {
			/* absolute branch address */
			ophex(cheat.tview.offset + fin->codeaddress, 8);
		    }
		};
		break;
	}
    }
}

static int legalmoderegister(long mode, long reg)
{
    int ret = 1;	/* default is ok. */

    switch(mode) {
	case 6:		/* Address register memory indirect with index  */
	    ret = 0;	
	    break;
	case 7:
	    if (3 == reg)	/* Program counter memory indirect with index */
		ret = 0;
	    if (5 == reg || 6 == reg || 7 == reg) 	/* Reserved */
		ret = 0;
	    break;
    }
    return ret;
}

static void moveinstruction(long kind)
{
    if (legalmoderegister(cheat.iview.srcmode, cheat.iview.srcreg)
	    && legalmoderegister(cheat.iview.extend, cheat.iview.destreg)) {
	opstring("move");
	sourcekind(kind == 1 ? 0 : (kind == 2 ? 2 : 1));
	padd();
	operand(cheat.iview.srcmode, cheat.iview.srcreg);
	opchar(',');
	operand(cheat.iview.extend, cheat.iview.destreg);
    } else
	invalid();
}


static void startrange(long bit)
{
    operand(bit <= 7 ? 0 : 1, bit % 8);
}


static void endrange(long first, long bit)
{
    if (first <= 7 && bit > 7) {
	endrange(first, 7);
	opchar('/');
	startrange(8);
	first = 8;
    };
    if (bit > first) {
	opchar('-');
	startrange(bit);
    }
}


static void registerlist(long kkkk, long mask)
{
    long bit = 0;
    long inrange = -1;
    long some = 0;
    while (bit <= 15) {
	if ( ( kkkk ? (mask >> (15 - bit)) : (mask >> bit) ) & 1 ) {
	    if (inrange < 0) {
		if (some)
		    opchar('/');
		startrange(bit);
		some = 1;
		inrange = bit;
	    }
	}
	else {
	    if (inrange >= 0) {
		endrange(inrange, bit - 1);
		inrange = -1;
	    }
	};
	bit++;
    };
    if (inrange >= 0)
	endrange(inrange, 15);
}


static void specialbits(void)
{
    long smod = cheat.iview.srcmode;
    long sreg = cheat.iview.srcreg;
    int  trapnum;

    switch (smod) {
	case 0:
	case 1:
	    trapnum = cheat.tview.offset & 0x000f;
	    if (15 == trapnum) 	{	/* PalmOS syscall special case */
		long	value;		/* syscall (TRAP #15) argument */
		char	*name;
		value = mynext(1);
		name = pilot_trap_name(value, fin->state.syscallname);
		opstring("SysTrap ");
		opstring(name);
	    } else {
		opstring("trap");
		padd();
		opchar('#');
		opint(trapnum);
	    }
	    break;
	case 2:
	    opstring("link");
	    padd();
	    operand(1, sreg);
	    opchar(',');
	    srckind = 1;
	    operand(7, 4);
	    break;
	case 3:
	    opstring("unlk");
	    padd();
	    operand(1, sreg);
	    break;
	case 4:
	    opstring("move");
	    padd();
	    operand(1, sreg);
	    opstring(",usp");
	    break;
	case 5:
	    opstring("move");
	    padd();
	    opstring("usp,");
	    operand(1, sreg);
	    break;
	case 6:
	    switch (sreg) {
		case 0:
		    opstring("reset");
		    break;
		case 1:
		    opstring("nop");
		    break;
		case 2:
		    opstring("stop");
		    padd();
		    srckind = 1;
		    operand(7, 4);
		    break;
		case 3:
		    opstring("rte");
		    break;
		case 4:
		    /*opstring("??rtd??");*/
		    invalid();
		    break;
		case 5:
		    opstring("rts");
		    break;
		case 6:
		    opstring("trapv");
		    break;
		case 7:
		    opstring("rtr");
		    break;
	    };
	    break;
	case 7:
	    invalid();
	    break;
    }
}


static void handlefour(void)
{
    long ext =  cheat.iview.extend;
    long dreg = cheat.iview.destreg;
    long smod = cheat.iview.srcmode;
    long sreg = cheat.iview.srcreg;
    long reglist;
    static char *unaryA[] = {"negx", "clr", "neg", "not", "", "tst"};
    static char *cross4[] = {"nbcd", "pea", "movem.w", "movem.l", "nbcd", 
	"swap", "ext.w", "ext.l"
    };
    static char *jumps[] = {"jsr", "jmp"};

    if (ext <= 2 && (dreg <= 3 || dreg == 5)) {
	opstring(unaryA[dreg]);
	sourcekind(ext);
	padd();
	operand(smod,sreg);
    }
    else if (dreg == 4 && ext <= 3) {
	opstring(cross4[smod == 0 ? ext + 4 : ext]);
	padd();
	if (ext >= 2 && smod != 0) {
	    registerlist(smod == 4, mynext(1));
	    opchar(',');
	};
	operand(smod, sreg);
    }
    else if (ext == 2 || ext == 3) {
	if (dreg == 6) {
	    opstring(cross4[ext]);
	    padd();
	    reglist = mynext(1);
	    operand(smod == 4 ? 3: smod, sreg);
	    opchar(',');
	    registerlist(0, reglist);
	}
	else if (dreg == 7) {
	    opstring(jumps[ext - 2]);
	    padd();
	    operand(smod,sreg);
	}
	else if (ext == 3) {
	    switch (dreg) {
		case 0:
		    opstring("move");
		    padd();
		    opstring("sr,");
		    operand(smod,sreg);
		    break;
		case 1:
		    invalid();
		    break;
		case 2:
		    opstring("move");
		    padd();
		    srckind=0;
		    operand(smod,sreg);
		    opstring(",ccr");
		    break;
		case 3:
		    opstring("move");
		    padd();
		    srckind=1;
		    operand(smod,sreg);
		    opstring(",sr");
		    break;
		case 5:
		    opstring("tas");
		    padd();
		    operand(smod,sreg);
		    break;
	    }
	}
    }
    else if (ext == 7) {
	opstring("lea");
	padd();
	operand(smod, sreg);
	opchar(',');
	operand(1, dreg);
    }
    else if (ext == 6) {
	opstring("chk");
	padd();
	srckind = 1;
	operand(smod,sreg);
	opchar(',');
	operand(0, dreg);
    }
    else if (ext == 1 && dreg == 7)
	specialbits();
    else
	invalid();
}


/* ...don't...even...ask! This code is magic, cursed, or both! --Perm */

/* Try filling out every field with at least 16 \0s, i.e. where there's more
 * than 0 of them, and less than 16 -> leave empty fields alone. */
static char *interpret[16] = {
    /*0*/ "\0\0\0\0\0\0\0\0",
    /*1*/ "",
    /*2*/ "",
    /*3*/ "",
    /*4*/ "\0\0\0\0\0\0\0\0",
    /*5*/ "\170\170\170\104\173\173\173\104addq\0s\0\0\0\0db\0\0\0subq\0\321\321\342",
    /*6*/ "\360\360\360\360\360\360\360\360b",
    /*7*/ "\200\200\200\200\200\200\200\200moveq",
    /*8*/ "\30\30\30\21\104\105\105\22or\0\0\0divu\0divs\0sbcd\0\70\23\143\0\0\70\0\0",
    /*9*/ "\30\30\30\50\102\102\102\50sub\0\0subx\0\70\31\151",
    /*A*/ "\0\0\0\0\0\0\0\0",
    /*B*/ "\30\30\30\50\103\103\103\50cmp\0\0eor\0\0cmpm\0\71\71\132",
    /*C*/ "\30\30\30\21\105\106\107\22and\0\0mulu\0muls\0abcd\0exg\0\0\70\23\143\0\0\70\64\224\0\0\70\0\64",
    /*D*/ "\30\30\30\50\102\102\102\50add\0\0addx\0\70\31\151",
    /*e*/ "",
    /*f*/ "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
};


/* Set the line-counter of the output function. */
void
disassemble_set_line(int line)
{
    if (line < 0 || line > DISASS_MAXYLINE)
	line = 0;
    fin->line = line;
}

/* Set whether to output hex opcodes or not (i.e. only memnonics). */
void
disassemble_set_detail(int full)
{
    fin->state.do_hexopcodes = full;
}

/* Set whether to output hex opcodes or not (i.e. only memnonics). */
void
disassemble_set_displaystate(int newstate)
{
    fin->state.display_state = newstate;
}

/* Set whether to output absolute or relative branch addresses */
void
disassemble_set_branchtype(int branchtype)
{
    fin->state.branch_type = branchtype;
}


/* Set whether to show system calls by their name or number. */
void
disassemble_set_syscall(int state)
{
    fin->state.syscallname = state;
}

void get_hexified_word(unsigned char *buf_p)
{
    long v;

    v = nextcodeword();
    opchar(' ');
    if (fin->unreadable & 2) {
	*buf_p++ = '\x95';
	opstring("\x95\x95");
    } else {
	*buf_p++ = v >> 8;
	ophex((v >> 8), -2);
    }
    if (fin->unreadable & 1) {
	*buf_p++ = '\x95';
	opstring("\x95\x95");
    } else {
	*buf_p++ = v;
	ophex(v, -2);
    }
}

/* Depending on the display state, either print a hex-ascii line, or a pure
 * ascii line. */
static void *
disassemble_single_hexline(void *addr)
{
    int alen, i;
    unsigned char buf[10], *buf_p = buf;

    fin->address = (unsigned char *)addr;
    fin->index_into = 0;
    fin->c = 0;
    fin->codeaddress = 0;
    fin->codesize = 0;
    fin->unreadable = 0;
    alen = 2;

    startline();			/* Initialize line output */
    lefthex((long)addr, 0, alen);	/* Print <alen> last digits of address */

    for (i = 0; i < 4; i++) {
	get_hexified_word(buf_p);
	buf_p += 2;
    }

    fin->address = buf;
    display_hex_line();		/* Alright, now display what we got. */

    /* Ok, pay attention: A hexline is _always_ 8 bytes, whether it failed
     * or not. So we _always_ return addr+8. */

    return (addr + 8);
}


/* Decode input data at address <addr> into a single instruction. No more. No
 * less.  Returns the address of the next instruction, or NULL on failure.
 */
void *
disassemble_single_line(void *addr)
{
    int alen;

    if (NULL == fin)
	return NULL;

    if (DISP_DISASM != fin->state.display_state)
	return disassemble_single_hexline(addr);
    
    fin->address = (unsigned char *)addr;
    fin->index_into = 0;
    fin->c = 0;
    fin->codeaddress = (long)addr;
    fin->codesize = 0;
    fin->unreadable = 0;
    alen = fin->state.do_hexopcodes ? 2 : 4;

    ErrTry {
	startline();			/* Initialize line output */
	lefthex((long)addr, 0, alen);	/* Print <alen> last digits of address */
	cheat.sview = mynext(1);	/* Grab first word */
	switch (cheat.iview.opcode) {	/* Aaaaaand it is, probably: */
	    case 0:
		handlezero();		/* opcode zero instructions */
		break;
	    case 1:
	    case 2:
	    case 3:			/* Some move-instruction */
		moveinstruction(cheat.iview.opcode);
		break;
	    case 4:
		handlefour();		/* opcode four instruction */
		break;
	    case 5:
	    case 6:
	    case 7:
	    case 8:
	    case 9:
	    case 10:
	    case 11:
	    case 12:
	    case 13:
	    case 15:
		breakfurther(interpret[cheat.iview.opcode]);
		break;
	    case 14:
		shiftinstruction();
		break;
	};
	displayline();			/* Alright, now display what we got. */

	return (void *)fin->codeaddress;/* Return address of next instruction */
    }

    /* If we got stuck with some unreadable address, we are teleported here. */
    ErrCatch(err) {
	if (err) {
	    opstring("**** Unreadable");
	    displayline();
	}
	return (void *)fin->codeaddress;
    } ErrEndCatch;
}

/* Initialize disassembler structures and allocate storage and whatnot.
 * Returns 0 if A-OK, or -1 on failure. */
int
initialize_disassembler(void)
{
    int ret = -1;

    fin = MemPtrNew(sizeof(disass_t));
    if (NULL != fin) {
	fin->opline = MemPtrNew(OPBUFSIZE+1);	/* magic */
	if (NULL != fin->opline) {
	    MemSet(fin->opline, OPBUFSIZE+1, 0);
	    fin->address = (unsigned char *)0;
	    fin->index_into = 0;
	    fin->c = 0;
	    fin->codeaddress = (long)0;
	    fin->codesize = 0;
	    fin->line = 0;
	    fin->unreadable = 0;
	    fin->state.do_hexopcodes = 1;
	    fin->state.display_state = DISP_DISASM;
	    ret = 0;
	} else {
	    MemPtrFree(fin);
	    fin = NULL;
	}
    }
    return ret;
}


/* Deallocates any storage allocated in initialize_disassembler(). */
void
terminate_disassembler(void)
{
    if (NULL != fin) {
	if (NULL != fin->opline) {
	    MemPtrFree(fin->opline);
	}
	MemPtrFree(fin);
	fin = NULL;
    }
}

