/* Main code for Disassembler.
 * (C) 1998 University Of Troms, All rights reserved.
 * All programming by Per Harald Myrvang <perm@stud.cs.uit.no>.
 */
/* The code now includes PalmOS3.0 StrPrintF workarounds.
 */

#include <Pilot.h>
#include <GraffitiReference.h>
#include <CharAttr.h>
#include "callback.h"

#include "disasmRsc.h"

#include "disasm.h"

#define PAGE_SIZE 24

struct prefstate_s {
    unsigned display:		1;	/* DISP_DISASM or DISP_HEX */
    unsigned detailed:		1;	/* with hexcodes in disasm */
    unsigned branchtype:	1;	/* 'bra $234789' or 'bra +12' */
    unsigned compacthex:	1;	/* '00005678' or '5678' OBSOLETE! */
    unsigned syscall_name:	1;	/* systrap number or name */
    unsigned enable_edit:	1;	/* allow memory edit or not */
    unsigned enable_stupid:	1;	/* allow user to do stupid things */
    unsigned can_search_again:	1;	/* if search_again can be done */
    unsigned debug:		1;	/* allow debug messages/resets */
};

struct searchstate_s {
    unsigned mode:	2;
    unsigned how:	2;
    unsigned where:	3;
};
enum searchmode { SM_HEX=0, SM_TEXT, SM_DIS };
enum searchhow { SH_CASE=0, SH_NOCASE, SH_GLOB, SH_REGEXP };
enum seachwhat { SW_REC=0, SW_DYN, SW_STO, SW_ROM, SW_ALL };

typedef struct {
    void		*address;		/* Position on exit */
    struct prefstate_s	state;
    /* To come: database-handle, record-handle, etc. */
    UInt current_db;		/* currently/last selected database, -1 for none */
    UInt current_rec;		/* currently/last selected record, -1 for none */
    int select_db_state;	/* 0 if databases, 1 if records */
    int current_reg;		/* currently/last selected registers, or -1. */
    void *searchaddr;		/* Start any searches here */
    unsigned long searchlen;	/* Search this many bytes */
    void *searchsegaddr;	/* Search segment starts here */
    unsigned long searchseglen;	/* Search segment length */
    void *search_lastmatch;	/* last match (needed for search-again) */
    struct searchstate_s searchstate;
#define MAX_SEARCH_WHAT 100
    int search_whatlen;		/* Length of search criterium (0=none) */
    unsigned char search_what[MAX_SEARCH_WHAT+1];/* Search criterium */
} DisassemblerPreferenceType;

static DisassemblerPreferenceType disasmPrefs;

/* Version number of preference-structure. Change when changing
 * DisassemblerPreferenceType. */
#define disasmPrefVersionNum  5

/* register list item structure */
typedef struct {
    char data[40];
    long value;
} regs_t;

typedef struct reclist {
    char name[50];
    void *ptr;
    long length;
} reclist_t;

static char *dbs_buf = NULL;

static char hexbuf[] = "0123456789ABCDEF";

static void *after_top_decode;
static long current_address;
static EventType fake_event;

/* Makes <n> digits out of <v>, and puts the result in <buf>. Will
 * zeroterminate <buf>. If <buf> is NULL, it will return a pointer to it's own
 * little buffer, which data will be valid until the next call. */
static char *
l2hex(register char *buf, register long v, register int n)
{
    static char mybuf[10];
    register char c, *s;

    if (8 < n)
	return NULL;
    s = (NULL == buf) ? mybuf : buf;
    while (n-- > 0)
	*s++ = (c = (v >> (n << 2)) & 0xF) <= 9 ? c + '0' : c - 10 + 'A';
    *s = '\0';
    return ((NULL == buf) ? mybuf : buf);
}

/* Ascii hexadecimal string to long. This one gives you enough rope to hang
 * yourself with, so be careful. */
static long h2l(unsigned char *s)
{
    int i, c;
    long l = 0;
 
    if (NULL == s)
	return 0;
    for (i = 0; '\0' != s[i]; i++) {
	c = s[i];
	if (c > 'F')	/* Ensure uppercase */
	    c -= 32;
	if (c > '9')	/* Ensure continous range over '0'-'F' */
	    c -= 7;
	c -= '0';	/* Value of hex digit */
	l = (l << 4) | (c & 0x0F);
    }
    return l;
}

static char *byte2hex(unsigned int i, char *buf)
{
    buf[0] = hexbuf[((i>>4) & 0xf)];
    buf[1] = hexbuf[(i & 0xf)];
    buf[2] = '\0';
    return buf;
}

static void set_object_visibility(Word id, int visible)
{
    FormPtr	frm;
    ControlPtr	ctr;

    frm = FrmGetActiveForm();
    ctr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, id));
    if (visible)
	CtlShowControl(ctr);
    else 
	CtlHideControl(ctr);
}

static void set_object_text(Int id, CharPtr new_text)
{
    FormPtr	frm;
    VoidPtr	fld;
    VoidHand	hnd;
    CharPtr	p;
    Word	i;
    FormObjectKind k;

    frm = FrmGetActiveForm();
    i = FrmGetObjectIndex(frm, id);
    k = FrmGetObjectType(frm, i);
    if (frmLabelObj == k) {
	FrmCopyLabel(frm, id, new_text);
    } else {
	fld = FrmGetObjectPtr(frm, i);
	FldSetSelection(fld, 0, 0);
	hnd = MemHandleNew(StrLen(new_text)+1);
	ErrFatalDisplayIf(hnd==NULL, "MemHandleNew");
	p = MemHandleLock(hnd);
	ErrFatalDisplayIf(p==NULL, "MemHandleLock");
	StrCopy(p, new_text);
	switch (k) {
	    case frmFieldObj:
		FldSetTextHandle(fld, (Handle)hnd);
		FldSetUsable(fld, true);
		FldDrawField(fld);
		break;
	    default:
		p = NULL;	/* i.e. illegal object type */
		break;
	}
	MemHandleUnlock(hnd);
	if (NULL == p) {
	    MemHandleFree(hnd);
	}
    }
}

static void set_num_field32(Word id, long num)
{
    char buf[10];

    StrPrintF(buf, "%lx", num);
    set_object_text(id, buf);
}

static void set_num_field16(Word id, unsigned int num)
{
    char buf[10];

    l2hex(buf, (long)num, 4);
    set_object_text(id, buf);
}

static void set_num_field16decparens(Word id, int num)
{
    char buf[10];

    StrPrintF(buf, "(%d)", num);
    set_object_text(id, buf);
}

static RectangleType dis_area = { {0, LISTING_START_Y+1}, {160, 160} };

static long dis(void *start)
{
    void *next;
    int i;

    WinEraseRectangle(&dis_area, 0);
    disassemble_set_line(0);
    after_top_decode = NULL;
    next = start;
    for (i = 0; i < 12; i++) {
	next = disassemble_single_line(next);
	if (NULL == after_top_decode)
	    after_top_decode = next;
    }
    set_num_field32(lblEndAddr, (long)next);
    return (long)next;
}

/* Instead of a macro. We go to great lengths to save space. */
static long dis_and_set_everyting(void)
{
    set_num_field32(fldAddr, (long)current_address);
    return dis((void *)current_address);
}

/* XXX: set_list() is a potenial gun, i.e. you may shot yourself in your foot
 * if you give it insane parameters.
 * TODO: Add proper checking of ALL parameters!
 */
static void
set_list(ListPtr lst, char **list, int num, int maxvis, int sel)
{
    int i;
    if ((NULL == lst) || (NULL == list))
	return;
    if (maxvis > 0) {	/* Yeah, yeah, nasty hack. So what. */
	i = (num > maxvis) ? maxvis : num;
	LstEraseList(lst);
    } else
	i = -maxvis;
    LstSetListChoices(lst, list, num);
    LstSetHeight(lst, i);
    LstSetSelection(lst, -1);	/* No selection */
    if ((sel < num) && (sel >= 0))
	LstMakeItemVisible(lst, sel);
    LstDrawList(lst);
    if ((sel < num) && (sel >= 0))
	LstSetSelection(lst, sel);
}

static void
set_popuplist(register FormPtr frm,
	register int listID,
	register int triggerID,
	register int selection)
{
    register ListPtr lst;
    register char *s;

    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID));
    LstSetSelection(lst, selection);
    s = LstGetSelectionText(lst, selection);
    if (NULL != s) {
	lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, triggerID));
	CtlSetLabel((ControlPtr)lst, s);
    }
}

static void PrefFormSetValue(FormPtr frm, Word id, Word value)
{
    ControlPtr ctr;
    ctr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, id));
    CtlSetValue(ctr, value);
}

static Boolean PrefFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    Word i;
    static struct prefstate_s state;

    CALLBACK_PROLOGUE;
    frm = FrmGetActiveForm();
    switch (e->eType) {
	case frmOpenEvent:
	    FrmDrawForm(frm);

	    i = (DISP_DISASM == disasmPrefs.state.display);
	    state = disasmPrefs.state;	/* read state */

	    i = state.branchtype ? btnRelative : btnAddress;
	    PrefFormSetValue(frm, i, 1);

	    i = state.syscall_name ? btnSysCallName : btnSysCallNum;
	    PrefFormSetValue(frm, i, 1);

	    i = state.enable_edit;
	    PrefFormSetValue(frm, chkMemEdit, i);

	    i = state.enable_stupid;
	    PrefFormSetValue(frm, chkAllowStupid, i);

	    i = state.debug;
	    PrefFormSetValue(frm, chkAllowDebug, i);

	    handled = true;
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case btnOk:
		    disassemble_set_branchtype(state.branchtype);
		    disassemble_set_syscall(state.syscall_name);
		    disasmPrefs.state = state;	/* save state */
		    /* tell main form to update itself. */
		    FrmUpdateForm(MainForm, UPDATE_LISTING);
		    /* FALLTHROUGH */
		case btnCancel:
		    FrmReturnToForm(MainForm);
		    handled = true;
		    break;
		case btnAddress:
		    state.branchtype = BRANCH_FULL;
		    break;
		case btnRelative:
		    state.branchtype = BRANCH_RELATIVE;
		    break;
		case btnSysCallNum:
		    state.syscall_name = 0;
		    break;
		case btnSysCallName:
		    state.syscall_name = 1;
		    break;
		case chkMemEdit:
		    state.enable_edit = e->data.ctlSelect.on;
		    break;
		case chkAllowStupid:
		    state.enable_stupid = e->data.ctlSelect.on;
		    break;
		case chkAllowDebug:
		    state.debug = e->data.ctlSelect.on;
		    break;
	    }
	    break;

	default:
	    break;
    }
    CALLBACK_EPILOGUE;
    return handled;
}


void print_current_db(int currentdb, int maxdbs, char **list)
{
    register int i;
    i = currentdb >= maxdbs ? -1 : currentdb;
    WinEraseChars(dbs_buf, StrLen(dbs_buf), 4, 12);
    StrPrintF(dbs_buf, "Current DB: %s", (i < 0) ? "<none>" : list[i]);
    WinDrawChars(dbs_buf, StrLen(dbs_buf), 4, 12);
}

static Boolean DBSelectFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    static char **database_list = NULL;
    static long database_num;
    static int state;
    static long recordaddress;
    static reclist_t **reclist = NULL;
    static UInt numrec = 0;
    static DmOpenRef dbref = NULL;
    LocalID dbid, rid;
    ListPtr lst;
    VoidHand v;
    UInt attrib;
    int i;
    int err = 0;

    CALLBACK_PROLOGUE;
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DBList));
    switch (e->eType) {
	case frmOpenEvent:
	    if (NULL == database_list) {
		/* Generate list of databases */
		database_num = DmNumDatabases(0);
		ErrFatalDisplayIf(0 == database_num, 
			"0 Databases?");
		i = ((database_num + 1) * sizeof(char *));
		database_list = MemPtrNew(i);
		ErrFatalDisplayIf(NULL == database_list,
			"no mem");
		MemSet(database_list, i, 0);
		for (i = 0; i < database_num; i++) {
		    database_list[i] = MemPtrNew(34);
		    ErrFatalDisplayIf(NULL == database_list[i],
			    "no mem");
		    dbid = DmGetDatabase(0, i);
		    ErrFatalDisplayIf(0 == dbid, 
			    "DmGetDatabase");
		    DmDatabaseInfo(0, dbid, database_list[i],
			    NULL, NULL, NULL, NULL, NULL,
			    NULL, NULL, NULL, NULL, NULL);
		    database_list[i][33] = '\0';
		}
	    }
	    /* If what the prefs think is the current db lies outside the
	     * domain of currently possible dbs, yank it to safe defaults. */
	    if (database_num < disasmPrefs.current_db) {
		disasmPrefs.select_db_state = 0;	/* Do dbs */
		disasmPrefs.current_db = database_num;	/* Last db */
		disasmPrefs.current_rec = -1;		/* First record */
	    }

	    FrmDrawForm(frm);
	    dbs_buf[0] = '\0';
	    dbref = NULL;
	    recordaddress = current_address;	/* default, i.e. no change */
	    print_current_db(-1, database_num, database_list);
	    state = disasmPrefs.select_db_state;
	    if (0 == state)
		FrmUpdateForm(DBSelectForm, SHOW_DB_LIST);
	    else
		FrmUpdateForm(DBSelectForm, SHOW_REC_LIST);
	    handled = true;
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;

	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case SHOW_DB_LIST:
		    FrmCopyTitle(frm, "Select database");
		    if (NULL == database_list)		/* Sanity check */
			break;
		    set_list(lst, database_list, database_num, 8,
			    disasmPrefs.current_db);
		    set_object_visibility(btnBack, 0);
		    handled = true;
		    break;
		case SHOW_REC_LIST:
		    FrmCopyTitle(frm, "Select record");
		    if (NULL == database_list)		/* Sanity check */
			break;
		    i = disasmPrefs.current_db;
		    print_current_db(i, database_num, database_list);

		    /* Generate list of records in database and show it */
		    dbid = DmGetDatabase(0, (UInt)i);
		    ErrFatalDisplayIf(0 == dbid,
			    "[REC] DmGetDatabase failed!");
		    DmDatabaseInfo(0, dbid, NULL, &attrib,
			    NULL, NULL, NULL, NULL, NULL,
			    NULL, NULL, NULL, NULL);
		    dbref = DmOpenDatabase(0, dbid,
			    dmModeReadOnly | dmModeShowSecret);
		    ErrFatalDisplayIf(NULL == dbref,
			    "DmOpenDatabase failed!");

		    if (attrib & dmHdrAttrResDB)
			numrec = DmNumResources(dbref);
		    else
			numrec = DmNumRecords(dbref);

		    i = ((numrec + 1) * sizeof(reclist_t *));
		    reclist = MemPtrNew(i);
		    ErrFatalDisplayIf(NULL == reclist,
			    "no mem");
		    MemSet(reclist, i, 0);
		    for (i = 0; i < numrec; i++) {
			reclist[i] = MemPtrNew(sizeof(reclist_t));
			if (disasmPrefs.state.debug) {
			    ErrFatalDisplayIf(NULL == reclist[i], "no mem");
			} else if (NULL == reclist[i]) {
			    err = 1; continue;
			}

			if (attrib & dmHdrAttrResDB)
			    DmResourceInfo(dbref, i, NULL, NULL, &rid);
			else 
			    DmRecordInfo(dbref, i, NULL, NULL, &rid);
			if (disasmPrefs.state.debug) {
			    ErrFatalDisplayIf(NULL == rid,
				    "[REC] no local id?");
			} else if (NULL == rid) {
			    err = 1; MemPtrFree(reclist[i]); reclist[i] = NULL; continue;
			}
				
			if (memIDHandle == MemLocalIDKind(rid)) {
			    v = MemLocalIDToGlobal(rid, 0);
			    if (disasmPrefs.state.debug) {
				ErrFatalDisplayIf(NULL == v,
					"[REC] no Local2Global");
			    } else if (NULL == v) {
				err = 1; MemPtrFree(reclist[i]); reclist[i] = NULL; continue;
			    }
			    reclist[i]->ptr = MemHandleLock(v);
			    MemHandleUnlock(v);
			} else
			    reclist[i]->ptr = MemLocalIDToGlobal(rid, 0);

			if (disasmPrefs.state.debug) {
			    ErrFatalDisplayIf(NULL == reclist[i]->ptr,
				    "[REC] No record pointer?");
			} else if (NULL == reclist[i]->ptr) {
			    err = 1; MemPtrFree(reclist[i]); reclist[i] = NULL; continue;
			}

			reclist[i]->length = MemPtrSize(reclist[i]->ptr);
			StrPrintF(reclist[i]->name, "#%d @ 0x%lx, %ld bytes",
				i, reclist[i]->ptr, reclist[i]->length);
		    } 
		    /* If what the prefs think is the current record lies
		     * outside the domain of currently possible records, yank
		     * it to a safe default. */
		    if (numrec < disasmPrefs.current_rec) {
			disasmPrefs.current_rec = 0;	/* First record */
		    }

		    if (err) {
			FrmCustomAlert(alertInfo, "An error occured!",
				"The database list and/or the list of records ",
				"may be inconsistent!");
		    }
		    
		    /* Yes, this works, because reclist is list of pointers to
		     * reclist_t entries, which begins with a string, which is
		     * what a list of pointers to strings are too, which is
		     * what LstSetListChoices() expects. Magic! Now get down on
		     * your knees and worship me. --Perm */
		    set_list(lst, (char **)reclist, numrec, 8, 
			    disasmPrefs.current_rec);
		    if (dbref != NULL) {
			DmCloseDatabase(dbref);
			dbref = NULL;
		    }
		    set_object_visibility(btnBack, 1);
		    state = 1;		/* enter show-records-state */
		    handled = true;
		    break;
	    }
	    break;

	case lstSelectEvent:
	    if (DBList == e->data.lstSelect.listID) {
		if (0 == state) {	/* i.e show-database state */
		    /* If we've choosen a new database, we reset the
		     * current-record index. */
		    if (disasmPrefs.current_db != e->data.lstSelect.selection)
			disasmPrefs.current_rec = -1;
		    /* Record selection id */
		    disasmPrefs.current_db = e->data.lstSelect.selection;
		    FrmUpdateForm(DBSelectForm, SHOW_REC_LIST);
		} else {
		    /* Record selection id */
		    i = e->data.lstSelect.selection;
		    disasmPrefs.current_rec = i;
		    if (NULL == reclist)		/* Sanity check */
			break;
		    current_address = (long)reclist[i]->ptr;
		    /* Set search data */
		    disasmPrefs.searchsegaddr = reclist[i]->ptr;
		    disasmPrefs.searchseglen = reclist[i]->length;
		    /* We leave by faking a cancel-button press, which will
		     * destroy the record list, and return to the main form.
		     * The cancel-button will also tell the main form to update
		     * itself. */
		    fake_event.eType = ctlSelectEvent;
		    fake_event.data.ctlSelect.controlID = btnCancel;
		    fake_event.data.ctlSelect.pControl = NULL;
		    EvtAddEventToQueue(&fake_event);
		}
		disasmPrefs.select_db_state = state;
		handled = true;
	    }
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case btnBack:
		    state = 0;
		    /* Destroy list of records here */
		    if (NULL != reclist) {
			for (i = 0; i < numrec; i++)
			    MemPtrFree(reclist[i]);
			MemPtrFree(reclist);
			reclist = NULL;
		    }
		    FrmUpdateForm(DBSelectForm, SHOW_DB_LIST);
		    handled = true;
		    break;

		case btnCancel:
		    /* Destroy list of records here */
		    if (NULL != reclist) {
			for (i = 0; i < numrec; i++)
			    MemPtrFree(reclist[i]);
			MemPtrFree(reclist);
			reclist = NULL;
		    }

		    /* Free any old stuff */
		    if (NULL != database_list) {
			for (i = 0; NULL != database_list[i]; i++)
			    MemPtrFree(database_list[i]);
			MemPtrFree(database_list);
			database_list = NULL;
		    }

		    /* Since we've possibly changed the search block, we set
		     * can_search_again to false, thus forcing a search within
		     * the [possibly] new block. */
		    disasmPrefs.state.can_search_again = 0;
		    
		    FrmReturnToForm(MainForm);
		    /* tell main form to update itself. */
		    FrmUpdateForm(MainForm, UPDATE_LISTING);
		    handled = true;
		    break;
	    }
	    break;

	case keyDownEvent:
	    switch(e->data.keyDown.chr) {
		case pageUpChr:
		case pageDownChr:
		    LstSetSelection(lst, -1);	/* No selection */
		    i = (pageUpChr == e->data.keyDown.chr) ? 0 : 1;
		    LstScrollList(lst, i, 7);
		    handled = true;
		    break;
	    }
	    break;

	default:
	    break;
    }
    CALLBACK_EPILOGUE;
    return handled;
}

/* Chip register stuff */

#define REGS_NUM 124

/* Indexes (relative from base) */
static int regs_index[REGS_NUM] = {
    0x000, 0x004, 0x100, 0x102, 0x104, 0x106, 0x108, 0x10A, 0x10C, 0x10E,
    0x110, 0x114, 0x118, 0x11C, 0x120, 0x124, 0x128, 0x12C, 0x130, 0x134,
    0x138, 0x13C, 0x140, 0x144, 0x148, 0x14C, 0x200, 0x202, 0x207, 0x300,
    0x302, 0x304, 0x308, 0x30C, 0x310, 0x400, 0x401, 0x403, 0x408, 0x409,
    0x40B, 0x410, 0x411, 0x413, 0x418, 0x419, 0x41A, 0x41C, 0x41D, 0x41F,
    0x420, 0x421, 0x422, 0x423, 0x428, 0x429, 0x42A, 0x42B, 0x430, 0x431,
    0x432, 0x433, 0x438, 0x439, 0x43B, 0x440, 0x441, 0x442, 0x443, 0x448,
    0x449, 0x44A, 0x44B, 0x500, 0x502, 0x504, 0x506, 0x600, 0x602, 0x604,
    0x606, 0x608, 0x60A, 0x60C, 0x60E, 0x610, 0x612, 0x614, 0x616, 0x618,
    0x61A, 0x61C, 0x700, 0x800, 0x802, 0x900, 0x902, 0x904, 0x906, 0x908,
    0xA00, 0xA05, 0xA08, 0xA0A, 0xA18, 0xA1A, 0xA1C, 0xA1F, 0xA20, 0xA21,
    0xA23, 0xA25, 0xA27, 0xA29, 0xA2B, 0xA2D, 0xA31, 0xA32, 0xB00, 0xB04,
    0xB0C, 0xB0E, 0xB10, 0xB12 
};


/* Value sizes, i.e. 1, 2, or 4 bytes */
static char regs_size[REGS_NUM+2] = 
    "\1\4\2\2\2\2\2\2\2\2"
    "\4\4\4\4\4\4\4\4\4\4"
    "\4\4\4\4\4\4\2\2\1\1"
    "\2\4\4\4\4\1\1\1\1\1"
    "\1\1\1\1\1\1\1\1\1\1"
    "\1\1\1\1\1\1\1\1\1\1"
    "\1\1\1\1\1\1\1\1\1\1"
    "\1\1\1\2\2\2\2\2\2\2"
    "\2\2\2\2\2\2\2\2\2\2"
    "\2\2\2\2\2\2\2\2\2\2"
    "\4\1\2\2\2\2\2\1\1\1"
    "\1\1\1\1\1\1\1\2\4\4"
    "\1\1\1\1"
    "\0\0";

static char *get_reg_name(int index)
{
    char *name;

/* The dataset is made by doing (simplified example - see Makefile):
    % perl -e '\
	print "sysCtrl\0mskRev\0csAGrpBase\0csBGrpBase\0csCGrpBase\0";' \
	> m68328Hwr_massaged.data
    % echo -ne "\0\0" >> m68328Hwr_massaged.data
    And then refering to it by hwrRegNames (id = 0x4001). Nasty, but it works.
*/
    name = make_trap_name((unsigned long)index, 0, hwrRegNames);
    if (NULL == name)
	name = "Unknown";
    return name;
}

static Boolean ChipRegsFormHandleEvent (EventPtr e)
{
    static regs_t **regs_list = NULL;
    Boolean handled = false;
    FormPtr frm;
    ListPtr lst;
    void *base;
    unsigned char *reg8;
    unsigned short *reg16;
    unsigned long *reg32;
    char buf[10];
    int i;

    CALLBACK_PROLOGUE;
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, RegsList));
    switch (e->eType) {
	case frmOpenEvent:
	    if (NULL == regs_list) {
		regs_list = MemPtrNew((REGS_NUM + 1) * sizeof(regs_t *));
		ErrFatalDisplayIf(NULL == regs_list, "no mem");
		for (i = 0; i < REGS_NUM; i++) {
		    regs_list[i] = MemPtrNew(sizeof(regs_t));
		    ErrFatalDisplayIf(NULL == regs_list[i],
			    "no mem");
		    MemSet(regs_list[i], sizeof(regs_t), 0);
		}
	    }
	    FrmDrawForm(frm);
	    FrmUpdateForm(ChipRegsForm, UPDATE_LISTING);
	    handled = true;
	    break;

	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case UPDATE_LISTING:
		    LstEraseList(lst);
		case REFRESH_LISTING:
		    if (NULL == regs_list)	/* Sanity check */
			break;
		    base = (void *)0xFFFFF000L;
		    if (UPDATE_LISTING == e->data.frmUpdate.updateCode)
			i = 0;
		    else 
			i = lst->topItem;
		    if ((i < 0) || (i >= REGS_NUM))	/* if bogus topItem */
			i = 0;
		    for (; i < REGS_NUM; i++) {
			if (NULL == regs_list[i])	/* sanity check */
			    break;
			reg8 = (unsigned char *)reg16 =
			    (unsigned short*)reg32 = (base + regs_index[i]);
			if ('\1' == regs_size[i]) {
			    regs_list[i]->value = *reg8 & 0xFFL;
			    byte2hex(regs_list[i]->value, buf);
			} else if ('\2' == regs_size[i]) {
			    regs_list[i]->value = *reg16 & 0xFFFFL;
			    l2hex(buf, (long)regs_list[i]->value, 4);
			} else if ('\4' == regs_size[i]) {
			    regs_list[i]->value = *reg32;
			    l2hex(buf, (long)regs_list[i]->value, 8);
			} else {
			    ErrDisplay("Register size isn't 1, 2 or 4!");
			    break;
			}
			StrPrintF(regs_list[i]->data, "0x%s  %s - 0x%s",
				l2hex(NULL, (long)regs_index[i], 4),
				get_reg_name(i), buf);
		    }
		    set_list(lst, (char **)regs_list, REGS_NUM, -8,
			    disasmPrefs.current_reg);
		    handled = true;
		    break;
	    }
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case btnOk:
		    FrmUpdateForm(ChipRegsForm, REFRESH_LISTING);
		    handled = true;
		    break;
		case btnCancel:
		    disasmPrefs.current_reg = lst->topItem;
		    if (NULL != regs_list) {
			for (i = 0; i < REGS_NUM; i++)
			    MemPtrFree(regs_list[i]);
			MemPtrFree(regs_list);
			regs_list = NULL;
		    }
		    FrmReturnToForm(MainForm);
		    handled = true;
		    break;
	    }
	    break;
	
	case keyDownEvent:
	    switch(e->data.keyDown.chr) {
		case pageUpChr:
		case pageDownChr:
		    LstSetSelection(lst, -1);	/* No selection */
		    i = (pageUpChr == e->data.keyDown.chr) ? 0 : 1;
		    LstScrollList(lst, i, 7);
		    disasmPrefs.current_reg = lst->topItem;
		    FrmUpdateForm(ChipRegsForm, REFRESH_LISTING);
		    handled = true;
		    break;
	    }
	    break;

	default:
	    break;
    }
    CALLBACK_EPILOGUE;
    return handled;
}

/* Memory editing Pre-commit table stuff */

#define PRECOMMITITEMS 16

/* These two are allocated by EditFormHandleEvent.OpenEvent, and freed by same
 * handler */
static unsigned int *precommit_value = NULL;	/* Pre-commit table */
static unsigned char *precommit_valid = NULL;	/* valid entries */

static void EditDrawCell(VoidPtr tbl, Word row, Word col, RectanglePtr bounds)
{
    char buf[10], *s;
    int currFont, datalen;
    unsigned long v;
    long l = 0;
    unsigned char *memptr, *itemptr;
    int i;

    CALLBACK_PROLOGUE;
    datalen = 0;
    memptr = (unsigned char *)TblGetRowData(tbl, row);	/* item (word) ptr */
    itemptr = (memptr + ((col - 1) * 2));			/* mem ptr */
    s = buf;
    v = 0;
    WinSetUnderlineMode(noUnderline);
    switch(col) {
	case 0 ... 4:		/* Hexadecimal data cells */
	    if (0 == col) {	/* column 0 special case */
		datalen = 2;
		byte2hex((unsigned int)((long)memptr & 0xff), s);
	    } else {
		datalen = 4;
		i = (row * 4) + (col - 1);
		/* See if we have a valid edited version of this cell. */
		if (PRECOMMITITEMS > i && precommit_valid[i]) {
		    v = (unsigned long)precommit_value[i];
		    l2hex(s, (long)v, 4);
		    // StrPrintF(s, "%x", (unsigned int)v);
		    s += 4;	/* 4 digits */
		    WinSetUnderlineMode(solidUnderline);
		}
		else
		    for(i = 0; i < 2; i++) {
			l = Read_Byte((void *)itemptr + i);
			v = (v << 8) | (l & 0xff);
			if (0xff < l) {
			    *s++ = '\x95';
			    *s++ = '\x95';
			} else {
			    byte2hex((unsigned int)l, s);
			    s += 2;
			}
		}
		*s = '\0';
		TblSetItemInt(tbl, row, col, v);
	    }
	    break;
	case 5:			/* character data cell */
	    datalen = 8;
	    for(i = 0; i < datalen; i++) {
		l = Read_Byte((void *)memptr + i);
		*s++ = (0xff < l) ? '\x95' : (unsigned char)l;
	    }
	    /* If there's valid edited cells for this row: integrate */
	    for(i = 0; i < 4; i++) {
		if (precommit_valid[(row * 4) + i]) {
		    ((unsigned int *)buf)[i] = precommit_value[(row * 4) + i];
		}
	    }
	    *s = '\0';
	    break;
	default:
	    datalen = 0;
	    break;
    }
    if (0 != datalen) {
	currFont = FntSetFont(0);
	WinEraseRectangle(bounds, 0);
	WinDrawChars(buf, datalen, bounds->topLeft.x + 1, bounds->topLeft.y);
	FntSetFont(currFont);
    }
    CALLBACK_EPILOGUE;
}

static Boolean EditFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    TablePtr tbl;
    FormPtr frm;
    FieldPtr afld, vfld, ffld;	/* Ptr to fldAddr, fldValue, & field w/focus. */
    static Word row, col;
    register Word numRows, numCols, focus;
    register WordPtr chattr;
    register char *s;
    unsigned int *dest, ui;
    register int i, j;
    static long addr, startaddr;
    static int is_editing;

    CALLBACK_PROLOGUE;
    frm = FrmGetActiveForm();
    tbl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, tblEdit));
    afld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fldAddr));
    vfld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fldValue));
    focus = FrmGetFocus(frm);		/* get index of field with focus */
    if (-1 != focus) {
	ffld = FrmGetObjectPtr(frm, focus);	/* Get pointer */
	focus = FrmGetObjectId(frm, focus);	/* focus is now field ID */
    } else
	ffld = NULL;
    chattr = GetCharAttr();
    switch (e->eType) {
	case frmOpenEvent:
	    /* Allocate memory for precommit table[s] */
	    ErrFatalDisplayIf((precommit_value || precommit_valid),
		    "Precommit tables already allocated at frmOpenEvent!");
	    precommit_value = MemPtrNew(sizeof(int) * (PRECOMMITITEMS + 1));
	    precommit_valid = MemPtrNew(sizeof(char) * (PRECOMMITITEMS + 1));
	    ErrFatalDisplayIf(((precommit_value == NULL) || (precommit_valid == NULL)),
		    "Can't allocate precommit tables!");
	    /* Invalidate precommit table */
	    for (i = 0; i < PRECOMMITITEMS; i++)
		precommit_valid[i] = '\0';
	    is_editing = 0;
	    row = col = -1;
	    startaddr = addr = current_address;
	    FrmDrawForm(frm);
	    FrmUpdateForm(EditForm, UPDATE_LISTING);
	    set_object_visibility(btnOk, 0);
	    set_num_field16decparens(lblDecVal, 0);
	    handled = true;
	    break;

	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case UPDATE_LISTING:
		case REFRESH_LISTING:
		    numRows = TblGetNumberOfRows(tbl);
		    addr = startaddr;
		    set_num_field32(fldAddr, addr);
		    numCols = 6;
		    for (i = 0; i < numRows; i++) {
			TblSetRowData(tbl, i, addr);
			addr += 8;
			for(j = 0; j < numCols; j++) {
			    TblSetItemStyle(tbl, i, j, customTableItem);
			}
			TblSetRowUsable(tbl, i, true);
			TblMarkRowInvalid(tbl, i);
		    }

		    for(j = 0; j < numCols; j++) {
			TblSetCustomDrawProcedure(tbl, j, EditDrawCell);
			TblSetColumnUsable(tbl, j, true);
		    }

		    FrmDrawForm(frm);
		    handled = true;
		    break;
	    }
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;
	
	case tblEnterEvent:
	    switch(e->data.tblEnter.column) {
		/* Always swallow enter-events for columns 0 and 5, thus making
		 * it impossible to select them. */
		case 0:
		case 5:
		    FldReleaseFocus(vfld);
		    handled = true;
		    break;
		/* Swallow other enter-events only if edit_enable is false. */
		default:
		    if (!disasmPrefs.state.enable_edit)
			handled = true;
		    break;
	    }
	    break;
	    
	case tblSelectEvent:
	    row = e->data.tblSelect.row;
	    col = e->data.tblSelect.column;
	    if (col > 0 && col < 5) { 
		set_num_field16(fldValue, TblGetItemInt(tbl, row, col));
		set_num_field16decparens(lblDecVal, TblGetItemInt(tbl, row, col));
		FrmSetFocus(frm, FrmGetObjectIndex(frm, fldValue));
		is_editing = 1;
	    }
	    break;

	case keyDownEvent:
	    switch(e->data.keyDown.chr) {
		case pageUpChr:
		case pageDownChr:
		    if (!is_editing) {
			if (pageUpChr == e->data.keyDown.chr)
			    startaddr -= 8;
			else
			    startaddr += 8;
			FrmUpdateForm(EditForm, UPDATE_LISTING);
			handled = true;
		    }
		    break;
		case prevFieldChr:	/* leave the field? */
		case nextFieldChr:
		case '\n':
		    if (fldAddr == focus) {	/* we're in fldAddr */
			if (!is_editing) {	/* we're not editing */
			    s = FldGetTextPtr(afld);	/* update addr. */
			    if (NULL != s) {
				startaddr = (h2l(s) & 0xfffffffeL);/* force even */
				FrmUpdateForm(EditForm, UPDATE_LISTING);
			    }
			} else	/* allow no changes to fldAddr if editing. */
			    set_num_field32(fldAddr, startaddr);
		    } else if (fldValue == focus) {	/* we're in fldValue */
			if (is_editing) {		/* we're editing */
			    /* Store new value into pre-commit table */
			    s = FldGetTextPtr(vfld);
			    if (NULL != s) {
				if (row < 0 || col < 1 || row > 3 || col > 4)
				    break;
				i = (row * 4) + (col - 1);
				precommit_valid[i] = '\1';
				precommit_value[i] = (unsigned int)h2l(s);
				set_num_field16(fldValue, 
					(int)precommit_value[i]);
				set_num_field16decparens(lblDecVal,
					(int)precommit_value[i]);
				FldReleaseFocus(vfld);
				set_object_visibility(btnOk,
					disasmPrefs.state.enable_edit);
				FrmUpdateForm(EditForm, UPDATE_LISTING);
			    }
			} else
			    FldEraseField(vfld);
		    } else {
			break;		/* Not handled */
		    }
		    handled = true;
		    break;
		case '0' ... '9':	/* hexadecimal chars are ok */
		case 'A' ... 'F':
		case 'a' ... 'f':
		    /* Accept input characters if 1) they go to fldValue and
		     * we're editing, or 2) they go to fldAddr and we're NOT
		     * editing. */
		    if ((fldValue == focus && is_editing)
			    || (fldAddr == focus && !is_editing))
			break;
		    /* anything non-control-char-ish is ignored. */
		default:
		    if (!IsCntrl(chattr, e->data.keyDown.chr))
			handled = true;			/* Swallow event */
		    break;
	    }
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case btnOk:
		    /* Commit changes in pre-commit-table */
		    if (!(disasmPrefs.state.enable_edit && is_editing))
			break;
		    if (0 != FrmCustomAlert(alertConf,
				"This might cause total destruction!\n\n"\
				"Are you sure you want to risk being an idiot?",
				"", ""))
			break;
		    for(i = 0; i < PRECOMMITITEMS; i++) {
			if (precommit_valid[i]) {
			    dest = (unsigned int *)(startaddr + (i * 2));
			    ErrFatalDisplayIf((long)dest & 0x1L, "Odd address!");
			    ui =  precommit_value[i];
			    /* We only actually write if the AllowStupid flag
			     * is set together with the enable_edit flag! */
			    if (disasmPrefs.state.enable_stupid) {
				//MemSemaphoreReserve(1);
				/* And now for some assembly black magic */
				FrmCustomAlert(alertInfo, "Pre write", "", "");
				asm("
				movem.l	%%a0/%%d0, -(%%sp)	/* Save regs used */

				/* Turn OFF memory protection */
				lea 0xfffff000, %%a0		/* Chipbase*/
				and.l	#0x7ffffff7, 0x114(%%a0)
				and.l	#0x7ffffff7, 0x118(%%a0)
				
				move.l	%0, %%a0	/* dest to A0 */
				move.w	%1, %%d0	/* ui to D0 */
				move.b	%%d0, 1(%%a0)	/* Write low 8 bits */
				lsr.w	#8, %%d0	/* Shift to high 8 bits */
				move.b	%%d0, 0(%%a0)	/* Write high 8 bits */

				/* Turn ON memory protection */
				lea 0xfffff000, %%a0		/* Chipbase*/
				or.l	#0x0008, 0x114(%%a0)
				or.l	#0x0008, 0x118(%%a0)

				movem.l	(%%sp)+, %%a0/%%d0	/* Restore regs used */
				" : : "g" (dest), "g" (ui));

				FrmCustomAlert(alertInfo, "Post write", "", "");

				//*dest = ui;
				//MemSemaphoreRelease(0);
			    } else {
				StrPrintF(dbs_buf,
					"writing 0x%s to addr 0x%lx (i=%d)",
					l2hex(NULL, (long)ui, 4),
					(long)dest, i);
				FrmCustomAlert(alertInfo, "DEBUG\n", dbs_buf,
					"\n(No write actually performed)");
			    }
			}
		    }
		    /* FALLTHROUGH */
		case btnCancel:
		    /* Free memory allocated to precommit tables */
		    if (precommit_value) MemPtrFree(precommit_value);
		    if (precommit_valid) MemPtrFree(precommit_valid);
		    precommit_valid = NULL;
		    precommit_value = NULL;
		    FrmReturnToForm(MainForm);
		    FrmUpdateForm(MainForm, UPDATE_LISTING);
		    break;
	    }
	    handled = true;
	    break;
	

	default:
	    break;
    }
    CALLBACK_EPILOGUE;
    return handled;
}

/* Search for <what> in parameters given by disasmPrefs.search and friends.
 * Returns NULL if not found, or the address of the first hit.
 */
static void *
search(register char *what)
{
    register void *ptr;
    register long todo;
    register int len;
    register int i, match, found;
    long l;

    if (NULL == what)
	return NULL;
    len = StrLen(what);
    if (0 == len)
	return NULL;

    ptr = NULL;	/* Assume failure */

    /* Store search criterium in case of a later search-again */
    l = len;
    if (MAX_SEARCH_WHAT < l)
	l = MAX_SEARCH_WHAT;
    disasmPrefs.search_whatlen = l;
    StrNCopy(disasmPrefs.search_what, what, l);
    disasmPrefs.search_what[MAX_SEARCH_WHAT] = '\0';

    if (SM_HEX == disasmPrefs.searchstate.mode) {
	if (len & 1) {
	    FrmCustomAlert(alertInfo,
		    "Uneven number of digits!", "", "");
	    return NULL;
	}
	l = h2l(what);
	i = len >> 1;	/* This is our search length */
	l = l << (32 - (i << 3));	/* Move "active" data up to MSB */
	ptr = disasmPrefs.searchaddr;
	todo = disasmPrefs.searchlen;
	if (0 == todo || NULL == ptr) {
	    FrmCustomAlert(alertInfo,
		    "No current record!", "", "");
	    return NULL;
	}
	todo -= (todo % i);	/* Prevent searcing beyond searchlength */
	found = 0;
	/* TODO: Do a proper search instead of this stupid bruteforce thing! */
	while (todo--) {
	    match = MemCmp(&l, ptr, i);
	    if (0 == match) {
		found = 1;
		break;
	    }
	    ptr += i;
	}
	if (! found)
	    ptr = NULL;
	else {
	    /* Remember search match in case of a search-again. */
	    disasmPrefs.search_lastmatch = ptr;
	}
    }
    else if (SM_TEXT == disasmPrefs.searchstate.mode) {
	FrmCustomAlert(alertInfo, "not yet implemented!", "", "");
    }
    else if (SM_DIS == disasmPrefs.searchstate.mode) {
	FrmCustomAlert(alertInfo, "not yet implemented!", "", "");
    }

    return ptr;
}

static void *
search_again(void)
{
    register void *last, *start, *ret;
    register long len;

    ret = NULL;
    if (disasmPrefs.state.can_search_again) {
	last = disasmPrefs.search_lastmatch;
	start = disasmPrefs.searchsegaddr;
	len = disasmPrefs.searchseglen;
	if (0 != len) {
	    /* See if last match were within search segment */
	    if ((last >= start) && (last < (start+len))) {
		/* If so, set new search segment */
		disasmPrefs.searchaddr = last + 1;	/* +1 or we'd match at once */
		disasmPrefs.searchlen = len - (last - start);
		ret = search(disasmPrefs.search_what);
		if (NULL == ret) {
		    /* If search failed, restore search segment and stuff */
		    disasmPrefs.search_lastmatch = last;
		    disasmPrefs.searchaddr = start;
		    disasmPrefs.searchlen = len;
		}
	    }
	}
	if (NULL != ret) {
	    /* If the search was successfull, do neccessary updates */
	    current_address = (long)ret;
	    SndPlaySystemSound(sndConfirmation);
	    FrmUpdateForm(MainForm, UPDATE_LISTING);
	} else
	    FrmCustomAlert(alertInfo,
		    "No match!.", "", "");
    } else
	FrmCustomAlert(alertInfo,
		"Do a search first!", "", "");

    return ret;
}

static Boolean SearchFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    FieldPtr fld;
    WordPtr chattr;
    void *foundaddr;
    int i;

    CALLBACK_PROLOGUE;
    frm = FrmGetActiveForm();
    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fldSearch));
    chattr = GetCharAttr();

    /* If there's any search value, show search-button */
    set_object_visibility(btnOk, (0 != FldGetTextLength(fld)));

    switch (e->eType) {
	case frmOpenEvent:
	    FrmDrawForm(frm);

	    set_popuplist(frm, ModeList, ModeTrigger,
		    disasmPrefs.searchstate.mode);

	    set_popuplist(frm, HowList, HowTrigger,
		    disasmPrefs.searchstate.where);
	    /* Only show HowTrigger if we're not searching for hex values */
	    set_object_visibility(HowTrigger,
		    (SM_HEX != disasmPrefs.searchstate.mode));

	    set_popuplist(frm, WhereList, WhereTrigger,
		    disasmPrefs.searchstate.where);

	    if (0 != disasmPrefs.search_whatlen) {
		FldSetSelection(fld, 0, FldGetTextLength(fld));
		FldInsert(fld, disasmPrefs.search_what,
			disasmPrefs.search_whatlen);
		FldSetSelection(fld, 0, 0);
	    }

	    handled = true;
	    break;

	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case UPDATE_LISTING:
		case REFRESH_LISTING:
		    FrmDrawForm(frm);
		    handled = true;
		    break;
	    }
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;
	
	case keyDownEvent:
	    switch(e->data.keyDown.chr) {
		default:
		    /* Allow only hexadececimal and control characters through
		     * if in hex mode */
		    if (SM_HEX == disasmPrefs.searchstate.mode) {
			if (! IsHex(chattr, e->data.keyDown.chr) &&
				! IsCntrl(chattr, e->data.keyDown.chr)) {
			    handled = true;
			    break;
			}
		    }
		    /* We accept anything else in all other modes */
		    handled = false;
		    break;
	    }
	    break;

	case popSelectEvent:
	    i = e->data.popSelect.selection;
	    switch (e->data.popSelect.listID) {
		case ModeList:
		    disasmPrefs.searchstate.mode = i;
		    set_object_visibility(HowTrigger, (SM_HEX != i));
		    break;
		case HowList:
		    disasmPrefs.searchstate.how = i;
		    break;
		case WhereList:
		    disasmPrefs.searchstate.where = i;
		    break;
	    }
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case btnOk:
		    /* Perform the search, then if there was a match, set the
		     * disassembler start address, then ask the main form to
		     * update itself, thus pointing us neatly to the found
		     * text, if any. */
		    if (SW_REC == disasmPrefs.searchstate.where) {
			/* Reset search segment */
			disasmPrefs.searchaddr = disasmPrefs.searchsegaddr;
			disasmPrefs.searchlen = disasmPrefs.searchseglen;
		    }
		    foundaddr = search(FldGetTextPtr(fld));
		    if (NULL != foundaddr) {
			disasmPrefs.state.can_search_again = 1;
			current_address = (long)foundaddr;
			SndPlaySystemSound(sndConfirmation);
		    }
		    else 
			FrmCustomAlert(alertInfo, "No match!", "", "");

		    /* FALLTHROUGH */
		case btnCancel:
		    FrmReturnToForm(MainForm);
		    FrmUpdateForm(MainForm, UPDATE_LISTING);
		    handled = true;
		    break;
	    }
	    break;
	

	default:
	    break;
    }
    CALLBACK_EPILOGUE;
    return handled;
}

static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    FieldPtr fld;
    WordPtr chattr;
    char update_addr_field, *s;
    static long last_address = 0;
    static int first_time = 1;
    static enum events lastEType;

    CALLBACK_PROLOGUE;

    chattr = GetCharAttr();
    frm = FrmGetActiveForm();
    FrmSetFocus(frm, FrmGetObjectIndex(frm, fldAddr));
    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fldAddr));

    switch (e->eType) {
	case frmOpenEvent:
	    FrmDrawForm(frm);
	    WinDrawLine(0, LISTING_START_Y, 159, LISTING_START_Y);
	    if (first_time) {
		char *s;
		void *foo;

		first_time = 0;

		/* Set display state, and update control */
		disassemble_set_displaystate(disasmPrefs.state.display);
		foo = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MethodList));
		LstSetSelection(foo, disasmPrefs.state.display);
		s = LstGetSelectionText(foo, disasmPrefs.state.display);
		if (NULL != s) {
		    foo = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MethodTrigger));
		    CtlSetLabel(foo, s);
		}

		/* Set detail, and update checkbox */
		disassemble_set_detail(disasmPrefs.state.detailed);
		foo = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, chkFull));
		CtlSetValue(foo, disasmPrefs.state.detailed);
		if (DISP_DISASM == disasmPrefs.state.display)
		    CtlShowControl(foo);
		else 
		    CtlHideControl(foo);
		WinDrawLine(0, LISTING_START_Y, 159, LISTING_START_Y);

		/* Set other configs */
		disassemble_set_branchtype(disasmPrefs.state.branchtype);
		disassemble_set_syscall(disasmPrefs.state.syscall_name);

		/* Set address, and update field and listing. */
		current_address = (long)disasmPrefs.address;
		last_address = dis_and_set_everyting();
	    }
	    handled = true;
	    break;

	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case UPDATE_LISTING:
		    last_address = dis_and_set_everyting();
		    handled = true;
		    break;
	    }
	    break;

	case menuEvent:
	    switch(e->data.menu.itemID) {
		case itemMemSelectDB:
		    FrmPopupForm(DBSelectForm);
		    break;
		case itemMemViewChip:
		    FrmPopupForm(ChipRegsForm);
		    break;
		case itemMemEditMem:
		    FrmPopupForm(EditForm);
		    break;
		case itemMemSearchMem:
		    FrmPopupForm(SearchForm);
		    break;
		case itemMemSearchAgain:
		    search_again();
		    break;
		case itemEditUndo:
		    if (fld)
			FldUndo(fld);
		    break;
		case itemEditCut:
		    if (fld)
			FldCut(fld);
		    break;
		case itemEditCopy:
		    if (fld)
			FldCopy(fld);
		    break;
		case itemEditPaste:
		    if (fld)
			FldPaste(fld);
		    break;
		case itemEditSelectAll:
		    if (fld)
			FldSetSelection(fld, 0, FldGetTextLength(fld));
		    break;
		case itemEditKbd:
		    SysKeyboardDialog(kbdDefault);
		    break;
		case itemEditGraf:
		    SysGraffitiReferenceDialog(referenceDefault);
		    break;
		case itemOptPrefs:
		    FrmPopupForm(PrefForm);
		    break;
		case itemOptHelp:
		    FrmHelp(hlpHelp);
		    break;
		case itemOptCopying:
		    FrmHelp(hlpCopy);
		    break;
		case itemOptAbout:
		    FrmCustomAlert(alertInfo, "Disassembler v" VERSION " Beta. "
		    "Compiled " __DATE__ ", " __TIME__ ".\n"
		    "All programming by Per Harald Myrvang <perm@stud.cs.uit.no>. "
		    "Disassembler based on Bill Rogers' Amiga code.\n"
		    "\251 1998 University of Troms\370, Norway", "", "");
		    break;
	    }

	    MenuEraseStatus(NULL);
	    handled = true;
	    break;
	
	case popSelectEvent:
	    if (MethodList == e->data.popSelect.listID) {
		disasmPrefs.state.display = e->data.popSelect.selection;
		disassemble_set_displaystate(disasmPrefs.state.display);
		last_address = dis((void *)current_address);
		set_object_visibility(chkFull,
			(DISP_DISASM == e->data.popSelect.selection));
		WinDrawLine(0, LISTING_START_Y, 159, LISTING_START_Y);
		/* Must be false, otherwise the popupfield won't change. */
		handled = false;
	    }
	    break;

	case penUpEvent:
	    /* The next test is neccessary in order to filter out excess
	     * baggage (i.e. pen events) from other forms. */
	    if (!(penDownEvent == lastEType || penMoveEvent == lastEType))
		break;
	    /* This might need a bit of explaining. When we receive a
	     * penUpEvent, we check the start-stroke and end-stroke
	     * y-coordinate to see the length and direction of the stroke.
	     * After determing whether this was a page-up/down _stroke_ or a
	     * line-up/down _tap_, we fake an event presenting itself as a
	     * corresponding button-press. And voila!
	     */
	     if ((e->data.penUp.start.y > LISTING_START_Y)
		     && (e->data.penUp.start.y < 160)) {
		 int dy;

		 fake_event.eType = ctlRepeatEvent;
		 dy = e->data.penUp.end.y - e->data.penUp.start.y;
		 if (abs(dy) < 5) {		/* A tap? */
		     if (((160 / 2) + (LISTING_START_Y / 2)) < abs(e->data.penUp.end.y))
			 fake_event.data.ctlRepeat.controlID = btnDown;
		     else
			 fake_event.data.ctlRepeat.controlID = btnUp;
		 }
		 else if (abs(dy) > 30) {	/* A stroke? */
		     if (0 < dy)
			 fake_event.data.ctlRepeat.controlID = btnPgDn;
		     else
			 fake_event.data.ctlRepeat.controlID = btnPgUp;
		 } else {	/* it's nothing, */
		     break;	/* so ignore it! */
		 }
		 fake_event.data.ctlRepeat.pControl = NULL;
		 fake_event.data.ctlRepeat.time = 100;	/* Dunno what this should be. */
		 EvtAddEventToQueue (&fake_event);
		 handled = true;
	     }
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case chkFull:
		    disasmPrefs.state.detailed =
			CtlGetValue(e->data.ctlSelect.pControl);
		    disassemble_set_detail(disasmPrefs.state.detailed);
		    last_address = dis((void *)current_address);
		    handled = true;
		    break;
	    }
	    break;

	case ctlRepeatEvent:
	    update_addr_field = '\1';
	    switch(e->data.ctlRepeat.controlID) {
		case btnPgUp:
		    current_address -= (unsigned long)last_address -
			(unsigned long)current_address + 2;
		    break;
		case btnPgDn:
		    current_address = last_address;
		    break;
		case btnUp:
		    current_address -= abs((long)current_address -
			    (long)after_top_decode);
		    break;
		case btnDown:
		    current_address = (long)after_top_decode;
		    break;
		case btnLeft:
		    current_address -= 2;
		    break;
		case btnRight:
		    current_address += 2;
		    break;
		default:
		    update_addr_field = '\0';
		    break;
	    }
	    if (update_addr_field) {
		last_address = dis_and_set_everyting();
		handled = false; /* must be false if repeat is to work! */
	    }
	    break;
	case keyDownEvent:
	    switch(e->data.keyDown.chr) {
		case pageUpChr:
		case pageDownChr:
		    if (pageUpChr == e->data.keyDown.chr)
			current_address -= (unsigned long)last_address -
			    (unsigned long)current_address + 2;
		    else
			current_address = last_address;
		    last_address = dis_and_set_everyting();
		    handled = true;
		    break;
		    /* events that may cause us to leave the field */
		case prevFieldChr:
		case nextFieldChr:
		case '\n':
		    /* Leave field, start disassembly */
		    s = FldGetTextPtr(fld);
		    if (NULL != s) {
			current_address = (h2l(s)& 0xfffffffe);	/* force even */
			last_address = dis_and_set_everyting();
		    }
		    handled = true;
		    break;
		/* hexadecimal chars are ok */
		case '0' ... '9':
		case 'A' ... 'F':
		case 'a' ... 'f':
		    break;
		/* anything non-control-char-ish is ignored. */
		default:
		    if (!IsCntrl(chattr, e->data.keyDown.chr)) {
			handled = true;			/* Swallow event */
		    }
		    break;
	    }
	    break;
	default:
	    break;
    }

    lastEType = e->eType;

    CALLBACK_EPILOGUE;

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	    case MainForm:
		FrmSetEventHandler(frm, MainFormHandleEvent);
		break;
	    case PrefForm:
		FrmSetEventHandler(frm, PrefFormHandleEvent);
		break;
	    case DBSelectForm:
		FrmSetEventHandler(frm, DBSelectFormHandleEvent);
		break;
	    case ChipRegsForm:
		FrmSetEventHandler(frm, ChipRegsFormHandleEvent);
		break;
	    case EditForm:
		FrmSetEventHandler(frm, EditFormHandleEvent);
		break;
	    case SearchForm:
		FrmSetEventHandler(frm, SearchFormHandleEvent);
		break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    Word prefsize;
    int	i;

    CALLBACK_PROLOGUE;
    /* Fetch application preferences */
    i = PrefGetAppPreferences('DiAs', 0, &disasmPrefs, &prefsize, true);

    /* If no prefs were found, reset all values. */
    if (noPreferenceFound == i) {
	disasmPrefs.address = NULL;
	disasmPrefs.state.display = DISP_DISASM;
	disasmPrefs.state.detailed = 1;
	disasmPrefs.state.branchtype = BRANCH_FULL;
	disasmPrefs.state.syscall_name = 0;
	disasmPrefs.state.enable_edit = 0;
	disasmPrefs.state.enable_stupid = 0;
	disasmPrefs.state.debug = 1;
	disasmPrefs.state.can_search_again = 0;
	disasmPrefs.searchstate.mode = SM_HEX;
	disasmPrefs.searchstate.how = SH_CASE;
	disasmPrefs.searchstate.where = SW_REC;
	disasmPrefs.searchsegaddr = NULL;
	disasmPrefs.searchseglen = 0;
	disasmPrefs.search_whatlen = 0;
	MemSet(disasmPrefs.search_what, MAX_SEARCH_WHAT, 0);
    }
    disasmPrefs.state.enable_stupid = 0;	/* For our sanity's sake! */

    dbs_buf = MemPtrNew(100);
    ErrFatalDisplayIf(dbs_buf == NULL, "no mem");
 
    initialize_disassembler();

    CALLBACK_EPILOGUE;

    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    CALLBACK_PROLOGUE;
    disasmPrefs.address = (void *)current_address;

    PrefSetAppPreferences('DiAs', 0, disasmPrefVersionNum,
	    &disasmPrefs, sizeof(DisassemblerPreferenceType), true);

    if (dbs_buf)
	MemPtrFree(dbs_buf);
    dbs_buf = NULL;

    FrmSaveAllForms();
    FrmCloseAllForms();
    terminate_disassembler();
    CALLBACK_EPILOGUE;
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
