#include <Pilot.h>
#include <SysTraps.h>
#include "disasmRsc.h"

/* The systraps_massaged.data resource must have resource id 0x4000 (16384).
 * The dataset systraps_massaged.data is created by doing this:
 *   % perl -ne 'if (m/sysTrap(\S+)\s*,.*$$/) { print "$$1\0"; }' \
 *      < /usr/local/pilot/m68k-palmos-coff/include/PalmOS/System/SysTraps.h \
 *      > systraps_massaged.data
 *   % echo -ne "\0\0" >> systraps_massaged.data
 * --Perm
*/

/* The trapnames are now from PalmOS3.0 */

static char *trapname[];

#define TRAP_BASE	sysTrapBase
#define TRAP_NUM	842	/* Was 694 in 2.0! */

#define LIBTRAP_BASE	sysLibTrapBase
#define LIBTRAP_NUM	6

const static char *lib_trap_name[] = {
    "sysLibTrapName", "sysLibTrapOpen", "sysLibTrapClose", "sysLibTrapSleep",
    "sysLibTrapWake", "sysLibTrapCustom"
};

const char *
make_trap_number(unsigned long trap)
{
    static char buf[10];	/* For numerical trap defs */
    StrPrintF(buf,"#%d", (int)((trap - TRAP_BASE) & 0xffff));
    return buf;
}

const char *
make_trap_name(unsigned long trap, unsigned long base, int rID)
{
    register char *trapnames, *name = NULL;
    VoidHand resH;

    trap -= base;
    /* Get pointer to sysTrapNames resource */
    resH = DmGetResource(strRsc, rID);
    ErrFatalDisplayIf(resH == NULL, "Can't GetResource!");
    if (NULL != resH) {
	trapnames = (char *)MemHandleLock(resH);
	ErrFatalDisplayIf(trapnames == NULL, "Can't MemHandleLock!");
	if (NULL != trapnames) {
	    for (name = trapnames; (('\0' != *name) && (0 != trap)); trap--) {
		while(*name++);
	    }
	    if ((0 != trap) || ('\0' == *name))
		name = NULL;
	}
	MemHandleUnlock(resH);
	DmReleaseResource(resH);
    }
    return name;
}

const char *
pilot_trap_name(unsigned long trap, int namestate)
{
    char *name = NULL;

    /* If trap is outside systrap boundaries, */
    if (!namestate || (trap < TRAP_BASE) || (trap >= TRAP_BASE + TRAP_NUM)) {
	/* See if trap is within libtrap boundaries */
	if (namestate && (trap >= LIBTRAP_BASE)
		&& (trap < LIBTRAP_BASE + LIBTRAP_NUM)) {
	    return lib_trap_name[trap - LIBTRAP_BASE];
	}
	/* Make syscall number-only */
	return make_trap_number(trap);
    }

    name = make_trap_name(trap, TRAP_BASE, sysTrapNames);
    if (NULL == name)
	name = make_trap_number(trap);

    return name;
}

#if 0
const char *
pilot_trap_name(unsigned long trap, int namestate)
{
    static char buf[10];	/* For numerical trap defs */
#ifdef VERBOSE
    /* If trap is outside systrap boundaries, */
    if (!namestate || (trap < TRAP_BASE) || (trap >= TRAP_BASE + TRAP_NUM)) {
	/* See if trap is within libtrap boundaries */
	if (namestate && (trap >= LIBTRAP_BASE)
		&& (trap < LIBTRAP_BASE + LIBTRAP_NUM)) {
	    return lib_trap_name[trap - LIBTRAP_BASE];
	}
#endif
	/* Make syscall number-only */
	StrPrintF(buf,"#%d", (int)((trap - TRAP_BASE) & 0xffff));
	return buf;
#ifdef VERBOSE
    }
    return trapname[trap - TRAP_BASE];
#endif
}

#ifdef VERBOSE
 #include "pilot-trap.h"
#endif /* VERBOSE */
#endif /* 0 */
